(** Time-stamp: <modified the 13/04/2010 (at 18:10) by Erwan Jahier> *)
 
  
open Predef
open PredefEvalType
open PredefEvalConst
open SyntaxTree
open SyntaxTreeCore
open Eff
open Printf
open Lxm
open Errors




(******************************************************************************)
let finish_me msg = print_string ("\n\tXXX evalType.ml:"^msg^" ->  finish me!\n")
let rec (f : Eff.id_solver -> Eff.val_exp -> Eff.val_exp * Eff.type_ list) =
  fun id_solver ve ->
    let ve_core, tl =
      match ve.core with
        | CallByPosEff ({it=posop; src=lxm}, OperEff args) -> (
            let posop_opt, args, tl = 
              try eval_by_pos_type id_solver posop lxm args
              with EvalType_error msg -> 
                raise (Compile_error(lxm, "type error: "^msg))
            in
            let ve = 
              match posop_opt with
                | None -> CallByPosEff ({it=posop; src=lxm}, OperEff args)
                | Some posop -> CallByPosEff ({it=posop; src=lxm}, OperEff args)
            in
              ve, tl
          )
        | CallByNameEff ({it=nmop; src=lxm}, nmargs ) ->
            let nmargs, tl = 
              try eval_by_name_type id_solver nmop lxm nmargs
              with EvalConst_error msg ->
                raise (Compile_error(lxm, "\n*** can't eval constant: "^msg))
            in
              CallByNameEff ({it=nmop; src=lxm}, nmargs ), tl
    in
      { core = ve_core; typ = tl ; clk = ve.clk }, tl
      

and (eval_by_pos_type : Eff.id_solver -> Eff.by_pos_op -> Lxm.t -> 
      Eff.val_exp list -> 
      Eff.by_pos_op option (* For op that hold a val_exp, we return the modified op *)
      * Eff.val_exp list   (* The args with type info added *)
      * Eff.type_ list     (* The type of the val_exp "posop(args)" *)
    ) =
  fun id_solver posop lxm args ->
    match posop with
      | Eff.Predef (op,sargs) -> (
          let args, targs = List.split (List.map (f id_solver) args) in
          let tve = PredefEvalType.f op lxm sargs targs in
            None, args, tve
        )
      | Eff.CALL node_exp_eff -> 
          let lti = List.map (fun v -> v.var_type_eff) node_exp_eff.it.inlist_eff in
          let lto = List.map (fun v -> v.var_type_eff) node_exp_eff.it.outlist_eff in
          let args, t_argsl = List.split (List.map (f id_solver) args) in
          let t_args = List.flatten t_argsl in
          let llti = List.length lti and lt_args = List.length t_args in
          let tve =
            if llti <> lt_args then
              raise (EvalType_error(sprintf 
           "\n*** arity error: %d argument(s) are expected, whereas %d is/are provided"
                         llti lt_args))
            else
              (match UnifyType.f lti t_args with
                 | UnifyType.Equal -> lto
                 | UnifyType.Unif subst -> List.map (subst_type subst) lto
                 | UnifyType.Ko msg -> raise (Compile_error(lxm, msg))
              )
          in
            None, args, tve
                
      | Eff.IDENT id  -> (
          let tve = (* [id] migth be a constant, but also a variable *)
            try [(id_solver.id2var id lxm).var_type_eff]
            with
                Unknown_var(lxm, idref) -> 
                  try [Eff.type_of_const (id_solver.id2const id lxm)]
                  with Unknown_constant _ -> 
                    raise (Unknown_var(lxm, idref))                  
          in
            None, [], tve
        )
      | Eff.WITH(ve) ->
          let ve, tve = f id_solver ve in
            Some(Eff.WITH(ve)), [], tve
            
      | Eff.TUPLE ->
          let args, targs = List.split (List.map (f id_solver) args) in
            None, args, List.flatten targs

      | Eff.CONCAT -> (
          let args, targs = List.split (List.map (f id_solver) args) in
          let tve =
            match targs with
              | [[Array_type_eff (teff0, size0)]; [Array_type_eff (teff1, size1)]] ->
                  let teff = 
                    match UnifyType.f [teff0] [teff1] with
                      | UnifyType.Equal -> teff1
                      | UnifyType.Unif subst -> subst_type subst teff1
                      | UnifyType.Ko msg -> raise (Compile_error(lxm, msg))
                  in
                    [Array_type_eff (teff, size0+size1)]
              | _ -> 
                  raise(EvalType_error(sprintf "arity error: 2 expected instead of %d" 
                                         (List.length args)))
          in
            None, args, tve
        )
      | Eff.STRUCT_ACCESS (fid) ->
          assert (List.length args = 1);          
          let arg, targ = f id_solver (List.hd args) in
          let teff_field =
            match targ with
              | [Struct_type_eff (name, fl)] -> (
                  try fst (List.assoc fid fl)
                  with Not_found ->
                    raise (
                      PredefEvalType.EvalType_error
                        (Printf.sprintf "%s is not a field of struct %s"
                           (Ident.to_string fid)
                           (LicDump.string_of_type_eff4msg (List.hd targ))))
                )
              | [x] -> PredefEvalType.type_error [x] "struct type"
              | x -> PredefEvalType.arity_error x "1"
          in
            None, [arg], [teff_field]

      | Eff.ARRAY_ACCES(i) ->
          assert (List.length args = 1);          
          let arg, targ = f id_solver (List.hd args) in
          let sz, teff =
            match targ with
              | [Array_type_eff(teff_elt, size)] -> size, teff_elt
              | _ ->
                  let msg = 
                    "\n*** Type error: '"^(LicDump.string_of_type_eff_list4msg targ) ^
                      "' was expected to be an array"
                  in
                    raise (Compile_error(lxm, msg))
          in
          let _ = if ((i >= 0) && (i < sz)) then () else
            raise(
              EvalType_error(sprintf "array index %d out of bounds 0..%d" i (sz-1)))
          in
            None, [arg], [teff]

      | Eff.ARRAY_SLICE(sieff) ->
          assert (List.length args = 1);          
          let arg, targ = f id_solver (List.hd args) in
          let sz, teff_elt =
            match targ with
              | [Array_type_eff(teff_elt, size)] -> size, teff_elt
              | _ ->
                  raise (Compile_error(
                           lxm, "\n*** Type error: '" ^ 
                             (LicDump.string_of_type_eff_list targ) ^ 
                             "' was expected to be an array"))
          in
          let _ = if ((sieff.se_first >= 0) && (sieff.se_first < sz)) then () else
            raise(
              EvalType_error(sprintf "array index %d out of bounds 0..%d" 
                               sieff.se_first (sz-1)))
          in
          let _ = if ((sieff.se_last >= 0) && (sieff.se_last < sz)) then () else
            raise(
              EvalType_error(sprintf "array index %d out of bounds 0..%d" 
                               sieff.se_last (sz-1)))
          in
            None, [arg], [Array_type_eff(teff_elt, sieff.se_width)]

      | Eff.HAT(size,ceff) -> 
          let ceff, teff_list = f id_solver ceff in
          let tl = List.map (fun teff -> Array_type_eff(teff, size)) teff_list in
            Some(Eff.HAT(size,ceff)), [], tl
              
      | Eff.ARRAY(args) ->
          (* check that args are of the same type *)
          let args, targs = List.split (List.map (f id_solver) args) in
          let teff_elt =
            List.fold_left
              (fun acc teff ->
                 match acc with
                   | [] -> teff
                   | [sacc] -> if acc = teff then acc else
                       raise(EvalType_error(
                               "all array elements should be of the same type"))
                   | _ -> assert false
              )
              []
              targs
          in
          let tve =
            assert (List.length teff_elt = 1);
            [Array_type_eff(List.hd teff_elt, List.length args)]
          in
            Some(Eff.ARRAY(args)), [], tve

      | Eff.WHEN clk_exp -> (
          let args, targs = List.split (List.map (f id_solver) args) in
          let _ = match clk_exp with
            | Base -> ()
            | NamedClock( { it = (cc,cv) ; src = lxm }) -> 
                let vi = id_solver.id2var (Ident.to_idref cv) lxm in
                  (match vi.var_type_eff with
	             | Eff.Bool_type_eff
                     | Eff.Enum_type_eff _ -> ()
	             | teff  -> 
		         let msg = "the type of a clock cannot be " ^
                           (LicDump.string_of_type_eff4msg teff) 
		         in
		           raise(Compile_error(lxm,msg))
                  )
          in
            match targs with
              | [teff] -> None, args, teff
              | _ -> raise(EvalType_error("arity error (1 arg expected)"))
        )
      | Eff.ARROW
      | Eff.FBY -> (
          let args, targs = List.split (List.map (f id_solver) args) in
            match targs with
              | [init; teff] -> if init = teff then None, args, teff else 
                  raise(EvalType_error("type mismatch. "))
              | _ -> raise(EvalType_error("arity error (2 args expected)"))
        )
      | Eff.CURRENT 
      | Eff.PRE -> (
          let args, targs = List.split (List.map (f id_solver) args) in
            match targs with
              | [teff] -> None, args, teff
              | _ -> raise(EvalType_error("arity error (1 arg expected)"))
        )
      | Eff.MERGE _ -> finish_me "merge"; assert false



and (eval_by_name_type : Eff.id_solver -> Eff.by_name_op -> Lxm.t -> 
      (Ident.t Lxm.srcflagged * Eff.val_exp) list -> 
      (Ident.t Lxm.srcflagged * Eff.val_exp) list * Eff.type_ list) =
  fun id_solver namop lxm namargs -> 
    match namop with
      | Eff.STRUCT_anonymous -> 
          (* ??? comment faire ici pour recuperer son type ???
             il faut que je recherche  l'aide des noms de champs
             le type structure qui va bien !

             - creer une table [liste des noms de champs -> ident de type structure] ?
             - rajouter dans la table a sa creation une entree dont le nom
             est compos du nom des champs ?
          *)
          finish_me "anonymous struct not yet supported"; 
          assert false
            (*    failwith "Finish me: anonymous struct not yet supported" *)

      | Eff.STRUCT (pn,opid) -> 
          let struct_type = id_solver.id2type opid lxm in
          let namargs = 
            match struct_type with
              | Struct_type_eff(sn, fl) -> 
                  let namargs = 
                    List.map
                      (fun (fn,fv) -> 
                         let lxm = fn.src in
                         let (ft,fopt) = 
                           try List.assoc fn.it fl with Not_found -> 
                             let msg = "type error: bad field"^(Ident.to_string fn.it) in
                               raise (Compile_error(lxm, msg))
                         in
                           (* let's check the type of fv *)
                         let fv, fv_type = f id_solver fv in
                         let _fv_type = (* XXX ignored? *)
                           match UnifyType.f [ft] fv_type with
                             | UnifyType.Unif t -> t
                             | UnifyType.Equal -> ft
                             | UnifyType.Ko msg -> raise (Compile_error(lxm, msg))
                         in
                           (fn,fv)                                 
                      )
                      namargs
                  in
                    namargs
(*                     Struct_type_eff(sn, fl) *)
              | _ -> 
                  raise (Compile_error(lxm, "type error: a structure is expected"))
            
          in
            namargs, [struct_type]
